% FUNCTIONS FOR MUSICAL HARMONY THEORY
% written in ECLiPSe Prolog <http://eclipseclp.org/>

% (c) Frieder Stolzenburg <fstolzenburg@hs-harz.de> 2016

:- set_flag(prefer_rationals,on).
:- set_flag(print_depth,9999).
:- lib(listut).


%%% TUNINGS

% common 12-tone tunings
tuning(equal,Tuning) :-
	ntet(12,Tuning).
tuning(just,Tuning) :-
	ntet(12,Scale),
	maplist(farey(0.011),Scale,Tuning).
tuning(kirnberger,
	[1_1,25_24,9_8,6_5,5_4,4_3,45_32,3_2,25_16,5_3,16_9,15_8]).
tuning(overtonal,Tuning) :-
	ntet(12,Scale),
	maplist(overtonal(0.01),Scale,Tuning).
tuning(pythagorean,Tuning) :-
	ntet(12,Scale),
        Const is 2^float(1/12),
        Percent is (Const-1)/(Const+1),
	maplist(power(Percent,3),Scale,Tuning).
tuning(rational,Tuning) :-
	ntet(12,Scale),
	maplist(farey(0.01),Scale,Tuning).

% equal temperament with n semitones (n-tet)
ntet(N,Tuning) :-
	(param(N),
	 for(I,0,N-1),
	 foreach(Ratio,Tuning) do
		Ratio is 2^float(I/N)).

% overtone tuning with maximal error (and octave equivalence)
overtonal(Percent,Value,Ratio) :-
	between(1,999,1,N),
	Ratio is rnd(Value*2^N)/2^N,
	test(Ratio,Value,Percent),
	!.

% tuning for n tones with powers of given overtone (including reciprocals)
power(Percent,Overtone,Value,Ratio) :-
	between(0,999,1,N),
	(	Ratio is Overtone^N/2^rnd(N*log2(Overtone)-log2(Value))
	;	Ratio is 2^rnd(N*log2(Overtone)+log2(Value))/Overtone^N),
	test(Ratio,Value,Percent),
	!.

% rational tuning with smallest possible denominator (with Farey sequences)
farey(Percent,Value,Ratio) :-
	tree(Value,Ratio),
	test(Ratio,Value,Percent),
	!.


%%% HARMONICITY CANDIDATES

% table of all possible harmonies with and all values wrt. given tuning
% m	number of tones in the harmony
% {...} set of semitones wrt. 12-tet
% g	granularity
% h	periodicity
% log h logarithmic periodicity
% g*h	product
% Omega number not necessarily distinct prime factors of g*h
% tau	number of divisors of g*h
% phi	number of positive integers <= g*h that are relatively prime
% sigma sum of divisors
% G     gradus suavitatis value of g*h according to Euler
% alpha average power
% ... average of above values follows, then:
% sim  percentage similarity according to Gill & Purves
thetable :-
	tuning(rational,Tuning),
	(for(I,1,11), foreach(I,List) do true),
	subset(Rest,List),
	M is length(Rest)+1,
	length(Simple,9),
	average(Tuning,[0|Rest],Simple,Average),
	printf("%d {0",[M]),
	(foreach(I,Rest) do printf(",%d",[I])),
	printf("}",[]),
	(foreach(J,Simple) do
		(integer(J) ->
			printf(" %d",[J])
		;	J_ is float(J),
			printf(" %.3f",[J_]))),
	(foreach(K,Average) do
		printf(" %.3f",[K])),
	similarity(Tuning,[0|Rest],Similarity),
	printf(" %.6f",[Similarity]),
	(harmony(Name,[0|Rest])
	->	printf(" %w",[Name])
	;	true),
	nl,
	fail.
thetable.

% arithmetic average values for possible harmonicity parameters
average(Tuning,Semitones,Simple,Average) :-
	simple(Tuning,Semitones,0,Simple),
	length(Simple,Extra), 
	Semitones = [0|Rest],
	(param(Tuning,Semitones,Extra),
	 foreach(Inversion,Rest),
	 fromto(Simple,In,Out,Average_) do
		length(List,Extra),
		simple(Tuning,Semitones,Inversion,List),
		(foreach(X,In),
		 foreach(Y,List),
		 foreach(Z,Out) do
			Z is X+Y)),
	length(Semitones,Len),
	(param(Len),
	 foreach(U,Average_),
	 foreach(V,Average) do
		V is float(U/Len)).

% simple values for possible harmonicity parameters
simple(Tuning,Semitones,Offset,[Granularity,Periodicity,Harmonicity,Product,Omega,Tau,Phi,Sigma,Gradus|Extra]) :-
	(param(Tuning,Offset),
	 foreach(Pitch,Semitones),
	 foreach(Ratio,Set_),
	     fromto(1,NumIn,NumOut,Granularity_),
	     fromto(1,DenIn,DenOut,Periodicity_) do
		Pos is Pitch-Offset,
		freq(Tuning,Pos,Ratio),
		NumOut is lcm(NumIn,numerator(Ratio)),
		DenOut is lcm(DenIn,denominator(Ratio))),
	Ref is -Offset, freq(Tuning,Ref,Factor),
	Granularity is integer(Granularity_/Factor),
	Periodicity is integer(Periodicity_*Factor),
	Harmonicity is log2(Periodicity),
	Product is integer(Granularity*Periodicity),
	prim(Product,Omega,Tau,Phi,Sigma,Gradus),
	(Extra = [Power]
	->	(param(Granularity_),
		 foreach(Ratio,Set_),
		 foreach(Num,Set) do
			Num is integer(Granularity_/Ratio)),
		energy(Set,Power)
	;	true).

% average energy of autocorrelation function (all-isi)
energy(Set,Result) :-
	lcmall(Set,Max),
%	(param(Set),
%	 for(Tau,1,Max),
%	 fromto(0,In,Out,Result_) do
%		Out is In+alpha(Set,Tau)^2),
	Result_ is number((
		between(1,Max,1,Tau),
		between(1,Max,1,T),
		member(N1,Set),
		divides(N1,T),
		Tmp is T+Tau,
		member(N2,Set),
		divides(N2,Tmp),
		between(1,Max,1,T_),
		member(N1_,Set),
		divides(N1_,T_),
		Tmp_ is T_+Tau,
		member(N2_,Set),
		divides(N2_,Tmp_)				
	)),
	Result is Result_/Max^2.

alpha(Set,Tau,Result) :-
	lcmall(Set,Max),
	(param(Set,Tau),
	 for(T,1,Max),
	 fromto(0,In,Out,Result) do
		Out is In+g2(Set,T,Tau)).

g1(Set,T,Result) :-
	Result is number((
		member(N,Set),
		divides(N,T)
	)).

g2(Set,T,Tau,Result) :-
	Tmp is T+Tau,
	Result is g1(Set,T)*g1(Set,Tmp).

% percentage similarity according to Gill and Purves (2009)
similarity(Tuning,Semitones,Average) :-
	findall(Similarity,
	 (subset([Semitone1,Semitone2],Semitones),
	  Interval is Semitone2-Semitone1,
	  freq(Tuning,Interval,Ratio),
	  X is numerator(Ratio),
	  Y is denominator(Ratio),
	  Similarity is (X+Y-1)/(X*Y)
	 ),List),
	(List = []
	->	Average is 1.0
	;	(foreach(S,List),
		 fromto(0,In,Out,Sum) do
		    Out is In+S),
		Average is float(Sum/length(List))).

% relative frequency of semitone from given tuning 
freq(Tuning,Semitone,Ratio) :-
	length(Tuning,Len),
	Mod is Semitone mod Len,
	Ratio is nth0(Mod,Tuning)*2^(Semitone div Len).


%%% CHORDS AND SCALES

% named harmonies (given as list of semitones) wrt. 12 tones
harmony(Name,Semitones) :-
	semitones(Name,Semitones_),
	(Name =.. [_,K]
	->	inversion(12,0,K,Semitones_,Semitones)
	;	Semitones = Semitones_).

% generate k-th inversion of harmony wrt. n tones
inversion(_,K,K,Semitones,Semitones).
inversion(N,K1,K2,[0,Offset|Rest],Inversion) :-
	K1 =< length(Rest),
	invert(N,Offset,Rest,Inversion_),
	K_ is K1+1,
	inversion(N,K_,K2,[0|Inversion_],Inversion).

% generate next inversion of given harmony
invert(N,Offset,[],[Last]) :-
	Last is N-Offset.
invert(N,Offset,[In|Rest],[Out|List]) :-
	Out is In-Offset,
	invert(N,Offset,Rest,List).

% harmonies with given Name
semitones(unison,[0]).
semitones(minor_second,[0,1]).
semitones(major_second,[0,2]).
semitones(minor_third,[0,3]).
semitones(major_third,[0,4]).
semitones(perfect_fourth,[0,5]).
semitones(tritone,[0,6]).
semitones(perfect_fifth,[0,7]).
semitones(minor_sixth,[0,8]).
semitones(major_sixth,[0,9]).
semitones(minor_seventh,[0,10]).
semitones(major_seventh,[0,11]).
semitones(major(_),[0,4,7]).
semitones(suspended(_),[0,5,7]).
semitones(minor(_),[0,3,7]).
semitones(diminished(_),[0,3,6]).
semitones(augmented,[0,4,8]).
semitones(major2(_),[0,2,4,7]).
semitones(major4(_),[0,4,5,7]).
semitones(major6(_),[0,4,7,9]).
semitones(major7(_),[0,4,7,10]).
semitones(major8(_),[0,4,7,11]).
semitones(minor2(_),[0,2,3,7]).
semitones(minor4(_),[0,3,5,7]).
semitones(minor6(_),[0,3,7,9]).
semitones(minor8(_),[0,3,7,11]).
semitones(zero,[0,3,6,9]).
semitones(stack5(_),[0,2,7,9]).
semitones(major9(_),[0,2,4,7,11]).
semitones(minor9(_),[0,2,3,7,10]).
semitones(pentatonic(_),[0,2,4,7,9]).
semitones(hexatonic,[0,2,4,6,8,10]).
semitones(ionian,[0,2,4,5,7,9,11]).
semitones(dorian,[0,2,3,5,7,9,10]).
semitones(phrygian,[0,1,3,5,7,8,10]).
semitones(lydian,[0,2,4,6,7,9,11]).
semitones(mixolydian,[0,2,4,5,7,9,10]).
semitones(aeolian,[0,2,3,5,7,8,10]).
semitones(locrian,[0,1,3,5,6,8,10]).
semitones(diatonic(_),[0,2,4,5,7,9,11]).
semitones(harmonic(_),[0,2,3,5,7,8,11]).
semitones(melodic(_),[0,2,3,5,7,9,11]).
semitones(blues(_),[0,2,3,4,5,7,9,10]).
semitones(chromatic,[0,1,2,3,4,5,6,7,8,9,10,11]).

% data from Johnson-Laird et al. (2012)
% chord#, semitones, rating, error, roughness
data(triad,"1a",[0,16,19],1.667,0.13,7.27).
data(triad,"1b",[0,15,20],2.889,0.25,6.11).
data(triad,"1c",[0,17,21],2.741,0.25,9.92).
data(triad,"2a",[0,19,22],2.481,0.28,4.00).
data(triad,"2b",[0,15,29],4.444,0.28,13.93).
data(triad,"2c",[0,14,21],3.630,0.26,11.52).
data(triad,"3a",[0,16,22],2.630,0.26,7.60).
data(triad,"3b",[0,6,20],3.556,0.31,17.85).
data(triad,"3c",[0,14,18],4.000,0.23,19.40).
data(triad,"4a",[0,15,22],2.926,0.26,8.94).
data(triad,"4b",[0,7,21],3.333,0.29,8.96).
data(triad,"4c",[0,26,29],4.296,0.29,11.03).
data(triad,"5a",[0,15,19],2.407,0.25,9.72).
data(triad,"5b",[0,16,21],3.593,0.23,10.77).
data(triad,"5c",[0,17,20],3.481,0.20,18.38).
data(triad,"6a",[0,7,14],3.148,0.27,9.83).
data(triad,"6b",[0,7,17],2.852,0.22,13.40).
data(triad,"6c",[0,10,17],3.111,0.29,12.13).
data(triad,"7a",[0,19,23],3.370,0.33,10.60).
data(triad,"7b",[0,4,17],3.296,0.24,22.67).
data(triad,"7c",[0,1,20],4.741,0.22,36.97).
data(triad,"8a",[0,16,23],2.630,0.21,10.97).
data(triad,"8b",[0,7,20],3.704,0.28,19.94).
data(triad,"8c",[0,1,5],3.963,0.24,37.07).
data(triad,"9a",[0,15,18],3.889,0.30,22.14).
data(triad,"9b",[0,15,21],3.519,0.27,13.69).
data(triad,"9c",[0,18,21],3.667,0.27,16.83).
data(triad,"10a",[0,11,14],3.963,0.28,13.90).
data(triad,"10b",[0,9,22],4.148,0.30,22.14).
data(triad,"10c",[0,1,3],5.185,0.25,54.29).
data(triad,"11a",[0,18,22],5.148,0.23,17.46).
data(triad,"11b",[0,4,18],3.444,0.28,20.93).
data(triad,"11c",[0,14,20],5.296,0.23,18.74).
data(triad,"12a",[0,17,23],3.926,0.29,18.67).
data(triad,"12b",[0,6,19],3.704,0.21,17.66).
data(triad,"12c",[0,1,6],4.519,0.22,39.64).
data(triad,"13a",[0,14,17],3.481,0.23,19.02).
data(triad,"13b",[0,10,27],3.333,0.21,17.07).
data(triad,"13c",[0,9,19],3.222,0.26,7.51).
data(triad,"14a",[0,15,26],3.630,0.28,24.85).
data(triad,"14b",[0,1,10],5.185,0.21,40.30).
data(triad,"14c",[0,14,27],4.000,0.27,30.65).
data(triad,"15a",[0,11,18],4.815,0.22,26.39).
data(triad,"15b",[0,1,7],5.296,0.24,35.40).
data(triad,"15c",[0,5,18],5.333,0.16,38.53).
data(triad,"16",[0,16,20],5.259,0.23,16.07).
data(triad,"17a",[0,15,23],5.593,0.21,17.27).
data(triad,"17b",[0,8,21],5.815,0.20,25.87).
data(triad,"17c",[0,1,4],4.148,0.20,41.16).
data(triad,"18a",[0,20,23],5.630,0.15,21.64).
data(triad,"18b",[0,3,16],4.259,0.26,23.31).
data(triad,"18c",[0,1,9],5.296,0.23,39.95).
data(triad,"19a",[0,14,25],5.593,0.20,30.42).
data(triad,"19b",[0,11,25],5.852,0.25,29.10).
data(triad,"19c",[0,11,22],6.667,0.09,33.88).
data(tetrad,"1",[0,10,16,19],2.54,0.18,22.10).
data(tetrad,"2",[0,11,16,19],2.08,0.13,22.17).
data(tetrad,"3",[0,14,16,19],2.00,0.18,24.08).
data(tetrad,"4a",[0,14,16,23],2.03,0.15,27.62).
data(tetrad,"4b",[0,8,10,19],3.59,0.22,32.81).
data(tetrad,"5",[0,11,14,19],2.38,0.18,28.60).
data(tetrad,"6",[0,10,14,16],3.38,0.22,40.39).
data(tetrad,"7a",[0,9,16,19],2.46,0.17,14.41).
data(tetrad,"b",[0,10,15,19],2.46,0.18,19.78).
data(tetrad,"8a",[0,9,14,19],2.41,0.18,17.51).
data(tetrad,"8b",[0,10,15,17],3.16,0.20,31.94).
data(tetrad,"9a",[0,14,16,21],2.81,0.18,24.91).
data(tetrad,"9b",[0,10,14,19],3.22,0.20,18.87).
data(tetrad,"10a",[0,11,16,21],2.92,0.18,25.20).
data(tetrad,"10b",[0,14,15,19],3.11,0.20,39.16).
data(tetrad,"11a",[0,11,14,21],2.95,0.19,32.41).
data(tetrad,"11b",[0,9,10,19],4.54,0.20,38.63).
data(tetrad,"12",[0,10,16,21],3.35,0.23,33.32).
data(tetrad,"13a",[0,9,11,19],3.54,0.25,37.32).
data(tetrad,"13b",[0,10,14,15],3.19,0.24,49.11).
data(tetrad,"14",[0,10,16,17],3.49,0.21,41.92).
data(tetrad,"15a",[0,9,15,19],3.43,0.20,15.60).
data(tetrad,"15b",[0,10,15,18],3.14,0.20,30.63).
data(tetrad,"16a",[0,6,16,19],4.83,0.03,20.70).
data(tetrad,"16b",[0,10,13,18],4.76,0.20,38.15).
data(tetrad,"17",[0,4,11,18],4.78,0.19,27.09).
data(tetrad,"18",[0,6,11,19],4.84,0.20,28.22).
data(tetrad,"19",[0,10,15,21],4.49,0.21,28.49).
data(tetrad,"20",[0,3,14,21],4.81,0.18,36.45).
data(tetrad,"21",[0,11,15,19],5.08,0.20,25.23).
data(tetrad,"22",[0,11,16,20],4.54,0.18,26.69).
data(tetrad,"23a",[0,10,14,18],4.73,0.16,28.34).
data(tetrad,"23b",[0,10,16,20],4.73,0.18,23.29).
data(tetrad,"24",[0,9,15,18],3.65,0.20,29.08).
data(tetrad,"25",[0,4,15,19],4.78,0.19,20.56).
data(tetrad,"26a",[0,11,15,21],5.22,0.21,25.35).
data(tetrad,"26b",[0,10,13,16],4.57,0.20,29.23).
data(tetrad,"27",[0,7,15,18],4.86,0.15,26.78).
data(tetrad,"28",[0,9,18,22],5.46,0.14,29.92).
data(tetrad,"29",[0,4,11,15],4.92,0.20,29.98).
data(tetrad,"30",[0,15,23,26],4.73,0.18,30.20).
data(tetrad,"31",[0,7,10,18],4.81,0.19,31.83).
data(tetrad,"32",[0,3,11,18],5.38,0.19,31.88).
data(tetrad,"33",[0,4,10,15],4.84,0.18,28.49).
data(tetrad,"34",[0,4,10,18],4.97,0.20,28.73).
data(tetrad,"35",[0,6,11,17],5.57,0.13,36.93).
data(tetrad,"36",[0,10,11,21],5.95,0.18,57.43).
data(tetrad,"37",[0,10,20,21],6.43,0.11,61.83).
data(misc,"blues5",[0,3,10,14,17],0.0,0.0,0.0).
data(misc,"blues6",[0,3,7,10,14,17],0.0,0.0,0.0).

% table for above data (triad or tetrad)
thetable(Type) :-
	tuning(rational,Tuning),
	data(Type,Chord,[0|Rest],Rating,_,Roughness),
	printf("%s {0",[Chord]),
	(foreach(I,Rest) do printf(",%d",[I])),
	printf("} %.3f %.3f",[Rating,Roughness]),
	length(Simple,9),
	average(Tuning,[0|Rest],Simple,Average),
	(foreach(J,Simple) do
		(integer(J) ->
			printf(" %d",[J])
		;	J_ is float(J),
			printf(" %.3f",[J_]))),
	(foreach(K,Average) do
		printf(" %.3f",[K])),
	similarity(Tuning,[0|Rest],Similarity),
	printf(" %.6f",[Similarity]),
	nl,
	fail.
thetable(_).

% Turkish and Central African scales (steps in cent)
% Turkish <https://de.wikipedia.org/wiki/Maqam_(Musik)>
scale(rast,[0,204,384,498,702,906,1086]).
scale(nihavent,[0,204,294,498,702,792,996]).
scale(ussak,[0,180,294,498,702,792,996]).
scale(huseyni,[0,180,294,498,702,882,996]).
scale(hicaz,[0,90,384,498,702,882,996]).
scale(huzzam,[0,114,318,432,702,816,1020]).
% African
scale(mbeme,[0,220,420,510,650,710,940]).
scale(fon,[0,240,470,680,910]).
scale(bofi,[0,380,880]).
scale(oubangui,[0,110,230,450,690,960]).
scale(ghana,[0,230,500,700,990]).

% compute frequency ratios (scale) from cents and approximate by ratios
convert(Percent,Cents,Scale,Ratios) :-
	maplist(percent,Cents,Scale),
	maplist(farey(Percent),Scale,Ratios).


%%% MATHEMATICAL AUXILIARY FUNCTIONS

% approximate value by fraction with denominator n
approx(N,Value,Ratio) :-
	eval(N,M),
	Ratio is rnd(M*Value)/M.

% signed relative error of y wrt. x
err(Y,X,Error) :-
	Error is eval(Y)/eval(X)-1.

% compute float fractional part of x
frac(X,Z) :-
	eval(X,Y),
	Z is Y-fix(Y).

% logarithm to base 2
log2(X,Log) :-
	Log is ln(eval(X))/ln(2).

% transform percent into cent, i.e. 1/100 of a semitone, and vice versa
percent(Cent,Percent) :-
	nonvar(Percent),
	Cent is 1200*log2(1+eval(Percent)).
percent(Cent,Percent) :-
	nonvar(Cent),
	Percent is 2^float(eval(Cent)/1200).

% round to nearest integer
rnd(X,N) :-
	N is integer(fix(eval(X)+1/2)).

% test whether absolute error of y wrt. x is within bound (percent)
test(Y,X,Percent) :-
	abs(err(Y,X)) =< Percent.


%%% NUMBER-THEORETIC ALGORITHMS

% M is integer divisor of N
divides(M,N) :-
	N mod M =:= 0.

% compute least common multiple (lcm) of all integers in Set
lcmall(Set,Result) :-
	(foreach(N,Set),
	 fromto(1,In,Out,Result) do
		Out is lcm(In,N)).

% count Number of solutions for Test predicate
number(Test,_) :-
	setval(number,0),
	call(Test),
	incval(number),
	fail.
number(_,Number) :-
	getval(number,Number).

% continued fraction development including coefficients list
cont(L,P_,Q) :-
	eval(P_,P),
        cont(L,P,1,1,0,0,1,Q).

cont([R|L],A0,A1,X0,X1,Y0,Y1,Q) :-
        R is fix(A0/A1),
        A2 is A0-R*A1,
        X2 is X0-R*X1,
        Y2 is Y0-R*Y1,
        ( L=[], Q is abs(Y2/X2)
        ; A2 =\=0, cont(L,A1,A2,X1,X2,Y1,Y2,Q)).               

% rational approximation with always decreased error (Stern-Brocot tree)
tree(X,Y2) :-
	eval(X,Y1),
	L0 is fix(Y1),
	R0 is L0+1,
	Y0 is rnd(Y1)/1,
	tree(Y1,L0/1,_,R0/1,_,[Y0],[Y2|_]).

% procedure following Forisek (2007)
tree(_,L,L,R,R,Ys,Ys).
tree(X0,L1in/L2in,Lout,R1in/R2in,Rout,[Y0|Yin],Yout) :-
	X0 =\= Y0,
	((L1in+R1in)/(L2in+R2in) < X0
	->	X1 is X0-abs(Y0-X0),
		K is fix((L2in*X1-L1in)/(R1in-R2in*X1))+1,
		New1 is L1in+K*R1in,
		New2 is L2in+K*R2in,
		Y1 is New1/New2,
		tree(X0,New1/New2,Lout,R1in/R2in,Rout,[Y1,Y0|Yin],Yout)
	;	X1 is X0+abs(Y0-X0),
		K is fix((R1in-R2in*X1)/(L2in*X1-L1in))+1,
		New1 is R1in+K*L1in,
		New2 is R2in+K*L2in,
		Y1 is New1/New2,
		tree(X0,L1in/L2in,Lout,New1/New2,Rout,[Y1,Y0|Yin],Yout)).


%%% PRIME NUMBERS

% compute number of prime factors (big omega) and of divisors (tau function),
% (small) phi, and sigma (number of divisors) for given n, and Euler's gradus suavitatis
prim(N,Omega,Tau,Phi,Sigma,Gradus) :-
	eval(N,N_),
	factor(N_,2,0,[],Res),
	prim_(Res,Omega,Tau,Phi,Sigma_,Gradus),
	integer(Sigma_,Sigma).

prim_([],0,1,1,1,1).
prim_([P^K|List],Omega,Tau,Phi,Sigma,Gradus) :-
	prim_(List,Omega_,Tau_,Phi_,Sigma_,Gradus_),
	Omega is Omega_+K,
	Tau is (K+1)*Tau_,
	Phi is Phi_*P^(K-1)*(P-1),
	Sigma is Sigma_*(P^(K+1)-1)/(P-1),
	Gradus is Gradus_+K*(P-1).

% factor n into list of prime powers p^k
factor(N,P,K,List,Res) :-
	(0 is N mod P
	->	N_ is N div P,
		K_ is K+1,
		factor(N_,P,K_,List,Res)
	;	P_ is P+1,
		(K > 0
		->	List_ = [P^K|List]
		;	List_ = List),
		(N > 1
		->	factor(N,P_,0,List_,Res)
		; 	Res = List_)).

